#ifndef YERBACON_TARGET_HPP
#define YERBACON_TARGET_HPP

#include <vector>
#include <string_view>
#include <iostream>
#include <memory>
#include <sstream>

#include "../parsing/ParseComponents.hpp"

#define INCLUDECOMPONENT(X) virtual void on(const X& parseComponent) {}
class Target {
protected:
    std::stringstream output;
    INCLUDECOMPONENT(StandardComponents::Define);
    INCLUDECOMPONENT(StandardComponents::types::String)
    INCLUDECOMPONENT(StandardComponents::Reference);
    INCLUDECOMPONENT(StandardComponents::Class);
public:
    static shared_ptr<Target> forName(string_view name);

    #define with(X) if (id == typeid(X)) { this->on(reinterpret_cast<X&>(*component)); continue; }
    string transpileWithTree(const ParseTree& tree) {
        output.str(string());
        for (const unique_ptr<ParseComponent>& component: tree) {
            const type_info& id = (*component).getId();
            with(StandardComponents::Define);
            with(StandardComponents::types::String);
            with(StandardComponents::Reference);
            with(StandardComponents::Class);
        }
        return output.str();
    };
    #undef with
};
#undef INCLUDECOMPONENT

inline string transpile(const ParseTree& tree, const string_view& language) {
    return Target::forName(language)->transpileWithTree(tree);
}

#include "implementations/Lua.hpp"
#include "implementations/Js.hpp"
#include "implementations/Py.hpp"

enum LANGUAGE: signed short { NONE = -1, LUA, JS, PY };
constinit array<string_view, 3> languages { ".lua", ".js", ".py"};

shared_ptr<Target> Target::forName(string_view name) {
    LANGUAGE selected = NONE;
    for (unsigned int i = 0; i < languages.size(); ++i) {
        if (name == languages[i]) {
            selected = static_cast<LANGUAGE>(i);
            break;
        }
    }
    shared_ptr<Target> target;
    #define ADDTARGET(X) target = make_shared<X>(X());
    switch (selected) {
    #ifdef LUA_HPP
        case LUA: ADDTARGET(LUA_HPP); break;
    #endif
    #ifdef JS_HPP
        case JS: ADDTARGET(JS_HPP); break;
    #endif
    #ifdef PY_HPP
        case PY: ADDTARGET(PY_HPP); break;
    #endif
        case NONE:
        default: {
            if (not name.empty()) {
                cout << '"' << (char) toupper(name.at(1));
                name.remove_prefix(2); cout << name << "\" is not a valid target." << endl;
            } else cout << "No target was provided." << endl;
            exit(0);
        }
    }
    #undef ADDTARGET
    return target;
};

#endif //YERBACON_TARGET_HPP